//
// MessagePack for Java
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
//
package org.msgpack.value;

import java.util.Iterator;
import java.util.List;

/**
 * Immutable representation of MessagePack's Array type.
 *
 * MessagePack's Array type can represent sequence of values.
 */
public interface ImmutableArrayValue
        extends ArrayValue, ImmutableValue
{
    /**
     * Returns an iterator over elements.
     * Returned Iterator does not support {@code remove()} method since the value is immutable.
     */
    Iterator<Value> iterator();

    /**
     * Returns the value as {@code List}.
     * Returned List is immutable. It does not support {@code put()}, {@code clear()}, or other methods that modify the value.
     */
    List<Value> list();
}
